/*
 * 6854 FIQ handler code
 *
 * (C) 1998, 1999 Phil Blundell <philb@gnu.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/linkage.h>

#define SR1		0
#define SR2		4
#define CR1		0
#define CR2		4
#define RXFIFO		8
#define TXCONTINUE	8
#define TXTERMINATE	12

#define RDA		1
#define CTS		16
#define TXU		32
#define TDRA		64
	
#define AP		1
#define FV		2
#define IDLE		4
#define ABT		8
#define ERR		16
#define DCD		32
#define OVRN		64

#define AC		0x01
#define RIE		0x02
#define TIE		0x04
#define RDSR_MODE	0x08
#define TDSR_MODE	0x10
#define DISCONTINUE	0x20
#define RXRESET		0x40
#define TXRESET		0x80
	
#define PSE		0x01
#define TWOBYTE		0x02
#define FLAGIDLE	0x04
#define CLRRX		0x20
#define CLRTX		0x40
#define RTS		0x80

	.section .rodata,"",%progbits
	.globl	SYMBOL_NAME(ec_fiq_code), SYMBOL_NAME(ec_fiq_code_end)
	
SYMBOL_NAME_LABEL(ec_fiq_code)
	ldr	pc, 1f
1:	.word	SYMBOL_NAME(ec_fiq_receive)
SYMBOL_NAME_LABEL(ec_fiq_code_end)

	/*
	 * Look for an incoming frame and buffer it.
	 * R8 -> ADLC base
	 * R9 -> buffer pointer
	 * R10 -> buffer end
	 * R11 -> temporary
	 * R12 -> status
	 * R13 -> my station number
	 */
	.text
	.globl	SYMBOL_NAME(ec_fiq_receive)
SYMBOL_NAME_LABEL(ec_fiq_receive)
	ldrb	r11, [r8, #SR1]
	tst	r11, #RDA
	bne	process_rda
	ldrb	r11, [r8, #SR2]
	tst	r11, #AP
	bne	process_ap
	mov	r12, r11
	tst	r11, #(FV | ERR)
	bne	endframe
	mov	r11, #(PSE | TWOBYTE | CLRRX)
	strb	r11, [r8, #CR2]
	b	request_irq
	
process_ap:
	ldrb	r11, [r8, #RXFIFO]
	teq	r11, #0xff
	teqne	r11, r13
	streqb	r11, [r9], #1
	subeqs	pc, lr, #4
	/* The frame isn't for us.  NB we don't currently check the
	   network number (which isn't available until later) so we
	   could collide with another station.  Fix this later.  */
	mov	r11, #(DISCONTINUE | RIE)
	strb	r11, [r8, #CR1]
	subs	pc, lr, #4

process_rda:
	ldrb	r11, [r8, #RXFIFO]
	strb	r11, [r9], #1
	ldrb	r11, [r8, #RXFIFO]
	strb	r11, [r9], #1
	cmp	r9, r10
	sublts	pc, lr, #4
	/* We overran the buffer */
	mov	r11, #DISCONTINUE
	strb	r11, [r8, #CR1]
	mov	r11, #(PSE | TWOBYTE | CLRRX)
	strb	r11, [r8, #CR2]
	mov	r12, #0
	b	request_irq

endframe:
	mov	r11, #0
	strb	r11, [r8, #CR2]
	ldrb	r11, [r8, #SR1]
	tst	r11, #RDA
	ldrneb	r11, [r8, #RXFIFO]
	strneb	r11, [r9], #1
	mov	r11, #RXRESET
	strb	r11, [r8, #CR1]
	mov	r11, #(PSE | TWOBYTE | CLRRX | CLRTX | FLAGIDLE | RTS)
	strb	r11, [r8, #CR2]
	/* Trigger a FIQ downgrade interrupt. */
request_irq:
	ldr	r10, irqmask
	ldrb	r11, [r10]
	orr	r11, r11, #128
	strb	r11, [r10]
	subs	pc, lr, #4

irqmask:
	.word	0x03200018
	
	/*
	 * Send a frame
	 * R8 -> ADLC base
	 * R9 -> buffer pointer
	 * R10 -> buffer end
	 * R11 -> temporary
	 * R12 -> status
	 */
	
	.globl	SYMBOL_NAME(ec_fiq_send_frame)
SYMBOL_NAME_LABEL(ec_fiq_send_frame)
	ldrb	r11, [r8, #SR1]
	tst	r11, #TDRA
	beq	not_tdra
	ldrb	r11, [r9], #1
	teq	r9, r10
	beq	all_over
	strb	r11, [r8, #TXCONTINUE]
	ldrb	r11, [r9], #1
	teq	r9, r10
	strneb	r11, [r8, #TXCONTINUE]
	subnes	pc, lr, #4
all_over:
	/* Transmission was successful. */
	strb	r11, [r8, #TXTERMINATE]
	mov	r11, #(PSE | TWOBYTE)
	strb	r11, [r8, #CR2]
	mov	r11, #0
	strb	r11, [r8, #CR1]
	b	request_irq

	/* Something went wrong. */
not_tdra:
	mov	r12, r11
	mov	r11, #0
	strb	r11, [r8, #CR1]
	mov	r11, #(PSE | CLRTX | TWOBYTE)
	strb	r11, [r8, #CR2]
	b	request_irq
	
