/*
 * Support for A680 onboard WD33C93 SCSI
 *
 * Copyright (C) 1999 Phil Blundell
 */

#include <linux/types.h>
#include <linux/mm.h>
#include <linux/blk.h>
#include <linux/sched.h>
#include <linux/version.h>
#include <linux/init.h>

#include <asm/setup.h>
#include <asm/page.h>
#include <asm/pgtable.h>
#include <asm/spinlock.h>
#include <asm/irq.h>

#include "../../scsi/scsi.h"
#include "../../scsi/hosts.h"
#include "../../scsi/wd33c93.h"
#include "a680.h"

static struct Scsi_Host *instance = NULL;

struct proc_dir_entry 
proc_scsi_a680 = 
{
  PROC_SCSI_A680, 5, "a680",
  S_IFDIR | S_IRUGO | S_IXUGO, 2
};

static void 
a680_scsi_intr (int irq, void *handle, struct pt_regs *fp)
{
  unsigned long flags;
  
  spin_lock_irqsave(&io_request_lock, flags);
  wd33c93_intr (instance);
  spin_unlock_irqrestore(&io_request_lock, flags);
}

static int 
dma_setup (Scsi_Cmnd *cmd, int dir_in)
{
}

static void 
dma_stop (struct Scsi_Host *instance, Scsi_Cmnd *SCpnt, int status)
{
}

int __init 
a680_scsi_detect (Scsi_Host_Template *tpnt)
{
  tpnt->proc_dir = &proc_scsi_a680;
  tpnt->proc_info = &wd33c93_proc_info;
  
  instance = scsi_register (tpnt, sizeof (struct WD33C93_hostdata));
  instance->base = 0;
  instance->irq = IRQ_HARDDISK;
  
  wd33c93_init(instance, 0, dma_setup, dma_stop, WD33C93_FS_8_10);

  request_irq(instance->irq, a680_scsi_intr, 0, "wd33c93 (A680)", NULL);

  return 1;
}

int 
a680_scsi_release (struct Scsi_Host *instance)
{
  free_irq(instance->irq, NULL);
#ifdef MODULE
  wd33c93_release();
#endif

  return 0;
}

#ifdef MODULE

Scsi_Host_Template driver_template = A680_SCSI;

#include "scsi_module.c"

#endif
