/*
 * Debug program interface IOCTL handler
 * for NetWinder
 * Woody 1998 woody@netwinder.org
 * Modified 1999 for 2.2 kernel
 */
/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/ptrace.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/smp.h>
#include <linux/mm.h>
#include <linux/init.h>
#include <linux/miscdevice.h>

#include <asm/dec21285.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/leds.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#define VERSION$ "0.20"
/*****************************************************************************/
#include "nwdebug.h"

void nwdebchar(char c);
void nwdebstr(char *p);

int  nwprintk(const char *fmt, ...);

static int	open_nwdebug(struct inode* inodep, struct file* filep);
static int	release_nwdebug(struct inode* inodep, struct file* filep);
static int	nwdebug_ioctl(struct inode* inodep, struct file* filep, unsigned int cmd, unsigned long arg);

static const char nwdebug_name[] = "nwdebug";

static struct file_operations nwdebug_fops = {
	NULL,			/* lseek		*/
	NULL,			/* read			*/
	NULL,			/* write		*/
	NULL,			/* readdir		*/
	NULL,			/* select		*/
	nwdebug_ioctl,		/* ioctl		*/
	NULL,			/* mmap			*/
	open_nwdebug,		/* open			*/
	NULL,			/* flush		*/
	release_nwdebug,	/* release		*/
	NULL,			/* fsync		*/
	NULL,			/* fasync		*/
	NULL,			/* check_media_change	*/
	NULL			/* revaldate		*/
};


void nwdebchar(char c)
{
volatile char* p = (volatile char *)(PCIO_BASE+0x3F8);
int timeout = 0x100000;

	while(!(p[5]&0x40) && timeout)
		{
		timeout--;
		}
	p[0] = c;
}

void nwdebstr(char *p)
{
char c;

	while ((c=*p)!='\0')
	{
		if (c == '\n')
		{
			nwdebchar('\r');
		}
		nwdebchar(c);
		p++;
	}
}

extern unsigned int current_cpld;
int	get_CPLD(void)
{
    return(current_cpld);
}

void	set_CPLD(int bits, int mask)
{
unsigned long flags;

    spin_lock_irqsave(&gpio_lock, flags);
    cpld_modify(mask,bits);
    spin_unlock_irqrestore(&gpio_lock, flags);
}

int	get_gpio(void)
{
    return(gpio_read());
}

void	set_gpio(int bits, int mask)
{
//unsigned long flags;

//    spin_lock_irqsave(&gpio_lock, flags);
    gpio_modify_io(mask,bits);
//    spin_unlock_irqrestore(&gpio_lock, flags);
}

int	getLED(void)
{
int i;

    i = gpio_read();
    return(((i&0x02)>>1) | ((i&0x80)>>6));
}

void	setLED(int color)
{
	leds_event(led_amber_off);
	leds_event(led_red_off);
	leds_event(led_green_off);
    
	switch(color)
	{
	case 3:
	    leds_event(led_amber_on);
	    break;
	case 2:
	    leds_event(led_red_on);
	    break;
	case 1:
	    leds_event(led_green_on);
	    break;
	default:
	    break;
	}
}

//void __netwinder_text 
int nwprintk(const char *fmt, ...)
{
va_list args;
int i;
char *buf_end;
char buf[256];
long flags;

	save_flags(flags);
	cli();
	va_start(args, fmt);
	i = vsprintf(buf, fmt, args); /* hopefully i < sizeof(buf)-1 */
	buf_end = buf + i;
	va_end(args);
	restore_flags(flags);
	nwdebstr(buf);

	return(i);
}


static int open_nwdebug(struct inode* inode, struct file* file)
{
	MOD_INC_USE_COUNT;
	return 0;
}


static int release_nwdebug(struct inode* inode, struct file* file)
{
	MOD_DEC_USE_COUNT;
	return 0;
}

static int nwdebug_ioctl(struct inode* inode, struct file* file, unsigned int cmd, unsigned long arg)
{
int	i;
int	bits,mask;
//int	oldCPLD;

//	printk(KERN_DEBUG "nwdebug_ioctl: cmd = 0x%X, arg = 0x%X.\n",cmd, *(int*)arg);

	switch (cmd)
	{
		case CMD_WRITE_CPLD:
			//expect 1 integer - bits, we need to create mask=0x0E
			//to prevent touching flash_writeable bit (bit 0)
			if (get_user(i, (int*)arg))
			    return -EFAULT;

//			oldCPLD = get_CPLD();	//get current setting
			mask = 0x7FFFFFFE;	//w/o top bit to be affected
			bits = i & mask;	//7 bits to be set
			set_CPLD(bits, mask);
//			printk(KERN_DEBUG "nwdebug: Old CPLD 0x%X, new CPLD 0x%X.\n",
//			    oldCPLD, get_CPLD());
			break;
		    

		case CMD_READ_CPLD:
			//return 1 integer
			i=get_CPLD();
//			printk(KERN_DEBUG "nwdebug: Current CPLD 0x%X.\n",i);
			if (put_user(i, (int*)arg))
			    return -EFAULT;
			break;
		    

		case CMD_WRITE_GPIO:
			//expect 1 integer - bits
			// we need to create mask

			if (get_user(i, (int*)arg))
			    return -EFAULT;

			mask = 0xFFF;		//all bits to be affected
			bits = i & mask;	// bits to be set
			set_gpio(bits, mask);
			break;
		    

		case CMD_READ_GPIO:
			//return 1 integer
			i=get_gpio();
//			printk(KERN_DEBUG "nwdebug: Current GPIO 0x%X.\n",i);
			if (put_user(i, (int*)arg))
			    return -EFAULT;
			break;
		    

		case CMD_GET_LED:
			//return 1 integer
			i=getLED();
			if (put_user(i, (int*)arg))
			    return -EFAULT;
			break;


		case CMD_SET_LED:
			//expect 1 integer - LED state
			// (0=off, 1=green, 2=red, 3=yellow)
			if (get_user(i, (int*)arg))
			    return -EFAULT;

			setLED(i);
			break;

		default:
			return -EINVAL;
	}
	return 0;
}


static struct miscdevice nwdebug_miscdev=
{
	NW_DEBUG_MINOR,
	"nwdebug",
	&nwdebug_fops
};

void nwdebug_init(void)
{
    misc_register(&nwdebug_miscdev);
    printk(KERN_NOTICE "Netwinder debug driver version %s.\n", VERSION$);
    nwprintk("Netwinder debug driver version %s.\n", VERSION$);

}

EXPORT_SYMBOL(nwprintk);

#ifdef MODULE

int init_module(void)
{
	nwdebug_init();
	return 0;
}

void cleanup_module(void)
{
	misc_deregister(&nwdebug_miscdev);
}
#endif


